 /* 
     nanoLang parser

     Copyright 2015-2016 by Stephan Schulz, schulz@eprover.org.

     This code is released under the GNU General Public Licence
     Version 2.
 */

%{
   #include <stdio.h>
   #include <string.h>
   #include "ast.h"
   #include "lexemetable.h"
   
   extern int yyerror(const char* err);
   extern int yylex(void);
   extern FILE *yyin;  

   AST_p ast;

%}

%locations
%define parse.error verbose
%define api.value.type {AST_p}


%start prog

%token IDENT
%token STRINGLIT
%token INTLIT
%token INTEGER STRING
%token IF WHILE RETURN PRINT ELSE
%token EQ NEQ LT GT LEQ GEQ
%left  PLUS MINUS
%left  MULT DIV
%right UMINUS
%token OPENPAR CLOSEPAR
%token SEMICOLON COMA
%token OPENCURLY CLOSECURLY


%token ERROR

%%


prog: /* Nothing */ {$$=NULL; ast=$$;}
    | prog def {
          $$ = ASTAlloc2(prog, NULL, 0, $1, $2);
          ast = $$;}
;

def: vardef {$$=$1;}
   | fundef {$$=$1;}
;

vardef: type IDENT SEMICOLON {
   $$ = ASTAlloc2(vardef, NULL, 0, $1, $2);}
;

fundef: type IDENT OPENPAR params CLOSEPAR body {
   $$ = ASTAlloc(fundef, NULL, 0, $1, $2, $4, $6);}
;

type: STRING  {$$ = $1;}
    | INTEGER {$$ = $1;}
;

params: /* empty */ {$$ = ASTAlloc2(params, NULL,0,NULL,NULL);}
      | paramlist {$$ = ASTAlloc2(params, NULL,0,$1,NULL);}
;

paramlist: param {$$ = $1;}
         | paramlist COMA param {
            $$ = ASTAlloc2(paramlist, NULL, 0, $1, $3);}
;

param: type IDENT {$$ = ASTAlloc2(param, NULL, 0, $1, $2);}
;

body: OPENCURLY vardefs stmts CLOSECURLY {
   $$ = ASTAlloc2(body, NULL, 0, $2, $3);}
;

vardefs: /* empty */ {$$=NULL;}
       | vardefs vardef {
          if($1)
          {
             $$ = ASTAlloc2(vardefs, NULL, 0, $1, $2);
          }
          else
          {
             $$ = $2;
          }}
;


stmts: /* empty */ {$$=NULL;}
     | stmts stmt {
          if($1)
          {
             $$ = ASTAlloc2(stmts, NULL, 0, $1, $2);
          }
          else
          {
             $$ = $2;
          }}
;


stmt: while_stmt   {$$=$1;}
    | if_stmt      {$$=$1;}
    | ret_stmt     {$$=$1;}
    | print_stmt   {$$=$1;}
    | assign       {$$=$1;}
    | funcall_stmt {$$=$1;}
;


while_stmt: WHILE OPENPAR boolexpr CLOSEPAR body {
   $$=ASTAlloc2(while_stmt, NULL, 0, $3, $5);
 }
;

if_stmt: IF OPENPAR boolexpr CLOSEPAR body{
   $$=ASTAlloc2(if_stmt, NULL, 0, $3, $5);
 }
       | IF OPENPAR boolexpr CLOSEPAR body ELSE body{
          $$=ASTAlloc(if_stmt, NULL, 0, $3, $5, $7, NULL);
 }

;

ret_stmt: RETURN expr SEMICOLON {
   $$=ASTAlloc2(ret_stmt, NULL, 0, $2, NULL);
 }
;

 print_stmt: PRINT expr SEMICOLON {
    $$=ASTAlloc2(print_stmt, NULL, 0, $2, NULL);
 }
;


assign: IDENT EQ expr SEMICOLON{
   $$=ASTAlloc2(assign, NULL, 0, $1, $3);
 }
;

funcall_stmt: funcall SEMICOLON{
   $$ = $1;
 }
;

boolexpr: expr EQ expr  {$$=ASTAlloc2(t_EQ, NULL, 0, $1, $3);}
        | expr NEQ expr {$$=ASTAlloc2(t_NEQ, NULL, 0, $1, $3);}
        | expr LT expr  {$$=ASTAlloc2(t_LT, NULL, 0, $1, $3);}
        | expr GT expr  {$$=ASTAlloc2(t_GT, NULL, 0, $1, $3);} 
        | expr LEQ expr {$$=ASTAlloc2(t_LEQ, NULL, 0, $1, $3);}
        | expr GEQ expr {$$=ASTAlloc2(t_GEQ, NULL, 0, $1, $3);}
;

expr: funcall {$$ = $1;}
    | INTLIT  {$$ = $1;}
    | IDENT   {$$ = $1;}
    | STRINGLIT {$$ = $1;}
    | OPENPAR expr CLOSEPAR {$$ = $2;}
    | expr PLUS expr  {$$=ASTAlloc2(t_PLUS, NULL, 0, $1, $3);}
    | expr MINUS expr {$$=ASTAlloc2(t_MINUS, NULL, 0, $1, $3);}
    | expr MULT expr  {$$=ASTAlloc2(t_MULT, NULL, 0, $1, $3);}
    | expr DIV expr   {$$=ASTAlloc2(t_DIV, NULL, 0, $1, $3);}
    | MINUS expr %prec UMINUS {$$=ASTAlloc2(t_MINUS, NULL, 0, $2, NULL);} 
;

funcall: IDENT OPENPAR args CLOSEPAR {
   $$=ASTAlloc2(funcall, NULL, 0, $1, $3);
 }
;

args: /* empty */ {$$=NULL;}
    | arglist {
       $$ = $1;
      }
;

arglist: expr {$$ = $1;}
       |expr COMA arglist {
          $$=ASTAlloc2(arglist, NULL, 0, $1, $3);
 }
;



%%


int yyerror(const char* err)
{
   printf("%d:%d: %s\n", yylloc.first_line, yylloc.first_column, err);
   return 0;
}


int main (int argc, char* argv[])
{
  int i;
  int res;
  bool printdot   = false;
  bool printsexpr = false;

   ++argv, --argc;  /* skip over program name */

   if (argc > 0)
   {
      if(strcmp(argv[0], "--dot")==0)
      {
         printdot   = true;
         printsexpr = false;         
         ++argv, --argc;
      }
      else if(strcmp(argv[0], "--sexpr")==0)
      {
         printdot   = false;
         printsexpr = true;         
         ++argv, --argc;
      }
   }
   
   if ( argc > 0 )
   {
      yyin = fopen( argv[0], "r" );
   }
   else
   {
      yyin = stdin;
   }   

   res = yyparse();

   if(res == 0)
   {
      if(printdot)
      {
         DOTASTPrint(stdout, ast);
      }
      if(printsexpr)
      {
         SExprASTPrint(stdout, ast);
         printf("\n");
      }
      ASTFree(ast);
   }
   //printf("About to free Lextable\n");
   LexTableFree(&lextable);
   return res;
}
